@@ -19,9 +19,11 @@ import timeit
 
 import cpppo
 
+logging.basicConfig( **cpppo.log_cfg )
 log				= logging.getLogger()
 log_not				= 0
 #log.setLevel( logging.INFO )
+#log.setLevel( logging.DEBUG )
 
 def test_logging():
     # Test lazy log message evaluation, ensuring it is at least an order of
@@ -431,6 +433,84 @@ def test_regex():
         assert num == 14
         assert sta is None and machine.current.name == '3'
 
+    regex			= str('.*')
+    machine			= cpppo.regex( name=str('dot'), initial=regex, terminal=True )
+    data			= cpppo.dotdict()
+    with machine:
+        source			= cpppo.chainable( str('aaab1230xoxx\0') )
+        try:
+            for i,(m,s) in enumerate( machine.run( source=source, data=data )):
+                log.info( "%s #%3d -> %10.10s; next byte %3d: %-10.10r: %r", m.name_centered(),
+                          i, s, source.sent, source.peek(), data )
+        except cpppo.NonTerminal:
+            pass
+        assert machine.terminal
+        assert i == 14
+        assert source.sent == 13
+        if sys.version_info[0] < 3:
+            assert data.input.input.tostring()  == 'aaab1230xoxx\x00'
+        else:
+            assert data.input.input.tounicode() == 'aaab1230xoxx\x00'
+
+    regex			= str('[^xyz]*')
+    machine			= cpppo.regex( name=str('not_xyz'), initial=regex )
+    data			= cpppo.dotdict()
+    with machine:
+        source			= cpppo.chainable( str('aaab1230xoxx\0') )
+        try:
+            for i,(m,s) in enumerate( machine.run( source=source, data=data )):
+                log.info( "%s #%3d -> %10.10s; next byte %3d: %-10.10r: %r", m.name_centered(),
+                          i, s, source.sent, source.peek(), data )
+        except cpppo.NonTerminal:
+            pass
+        assert not machine.terminal
+        assert i == 9
+        assert source.sent == 8
+        if sys.version_info[0] < 3:
+            assert data.input.input.tostring()  == 'aaab1230'
+        else:
+            assert data.input.input.tounicode() == 'aaab1230'
+
+    regex			= str('[^\x00]*')
+    machine			= cpppo.regex( name=str('not_NUL'), initial=regex )
+    data			= cpppo.dotdict()
+    with machine:
+        source			= cpppo.chainable( str('aaab1230xoxx\0') )
+        for i,(m,s) in enumerate( machine.run( source=source, data=data )):
+            log.info( "%s #%3d -> %10.10s; next byte %3d: %-10.10r: %r", m.name_centered(),
+                      i, s, source.sent, source.peek(), data )
+        assert i == 13
+        assert source.sent == 12
+        if sys.version_info[0] < 3:
+            assert data.input.input.tostring()  == 'aaab1230xoxx'
+        else:
+            assert data.input.input.tounicode() == 'aaab1230xoxx'
+
+def test_regex_demo():
+    regex			= str( '(ab+)((,[ ]*)(ab+))*' )
+    machine			= cpppo.regex( name=str( 'demo' ), initial=regex )
+    data			= cpppo.dotdict()
+    with machine:
+        source			= cpppo.chainable( str( 'abbb, abb, ab' ))
+        for i,(m,s) in enumerate( machine.run( source=source, data=data )):
+            log.info( "%s #%3d -> %10.10s; next byte %3d: %-10.10r: %r", m.name_centered(),
+                      i, s, source.sent, source.peek(), data )
+        assert i == 14
+        assert source.sent == 13
+
+    regexstr, lego, machine, initial = cpppo.state_input.from_regex(
+            regex, alphabet=cpppo.type_str_iter, encoder=None, 
+            typecode=cpppo.type_str_array_symbol, context=None )
+    assert str( lego ) == "ab+(, *ab+)*"
+    assert str( machine ) == """\
+  name final?   , None a b 
+---------------------------
+* 0    False  1 1 1    2 1 
+  1    False  1 1 1    1 1 
+  2    False  1 1 1    1 3 
+  3    True   1 4 1    1 3 
+  4    False  4 1 1    2 1 
+"""
 
 import binascii
 import codecs
@@ -452,7 +532,9 @@ def test_codecs():
 
     # Test parsing of greenery.fsm/lego regexes specified in Unicode.  Then,
     # generate corresponding cpppo state machines that accept Unicode input
-    # symbols, and byte input symbols.
+    # symbols, and byte input symbols.  These tests will accept as much of the
+    # input as matches the regular expression.
+
 
     texts 			= [
         'pi: π',
@@ -460,51 +542,58 @@ def test_codecs():
         'This contains π,π and more πs',
         'a 480Ω resistor',
         ]
+    tests			= [
+        ('[^π]*(π[^π]*)+',	True),	# Optional non-π's, followed by at least one string of π and non-π's
+        ('[^π]*[^π]',		False) 	# Any number of non-π, ending in a non-π
+        ]
 
     for text in texts:
-        # First, convert the unicode regex to a state machine in unicode symbols.
-        with cpppo.regex( name='pies',  context="pies", initial='.*π.*', terminal=True ) as pies:
-            source			= cpppo.chainable( text )
-            data			= cpppo.dotdict()
-            try:
-                for mch, sta in pies.run( source=source, data=data ):
+        for re,tr in tests:
+            # First, convert the unicode regex to a state machine in unicode symbols.  Only if both
+            # the dfa and its sub-state are "terminal", will it be terminal.
+            with cpppo.regex(
+                    name='pies',  context="pies", initial=re, terminal=True ) as pies:
+                original		= text
+                source			= cpppo.chainable( original )
+                data			= cpppo.dotdict()
+                try:
+                    for mch, sta in pies.run( source=source, data=data ):
+                        pass
+                except cpppo.NonTerminal:
                     pass
-            except cpppo.NonTerminal:
-                pass
-                    
-            log.info( "%s ends: %s: %r", pies.name_centered(),
-                      "string accepted" if pies.terminal else "string rejected", data )
-        
-            # Each of these are greedy, and so run 'til the end of input (next state
-            # is None); they collect the full input string.
-            assert pies.terminal == ( 'π' in text )
-            assert data.pies.input.tounicode() == text
+                accepted		= pies.terminal and data.pies.input.tounicode() == original
+                log.info( "%s ends w/ re %s: %s: %r", pies.name_centered(), re,
+                          "string accepted" if accepted else "string rejected", data )
+            
+                # Each of these are greedy, and so run 'til the end of input (next state is None); they
+                # collect the full input string, unless they run into a non-matching input.
+                expected		= tr == ('π' in text )
+                assert accepted == expected
 
     for text in texts:
         # Then convert the unicode regex to a state machine in bytes symbols.
         # Our encoder generates 1 or more bytes for each unicode symbol.
+        for re,tr in tests:
+            original		= text.encode( 'utf-8' ) # u'...' --> b'...'
+            source		= cpppo.chainable( original )
+            data		= cpppo.dotdict()
 
-        pies			= cpppo.regex( name='pies', context="pies", initial='.*π.*',
-                                               terminal=True,
-                                               regex_alphabet=int,
-                                               regex_typecode='B',
-                                               regex_encoder=lambda s: ( b for b in s.encode( 'utf-8' )))
-        
-        source			= cpppo.chainable( text.encode( 'utf-8' ))
-        data			= cpppo.dotdict()
-
-        with pies:
-            try:
-                for mch, sta in pies.run( source=source, data=data ):
+            with cpppo.regex(
+                    name='pies', context="pies", initial=re, terminal=True,
+                    regex_alphabet=int,
+                    regex_typecode='B',
+                    regex_encoder=lambda s: ( b for b in s.encode( 'utf-8' ))) as pies:
+                try:
+                    for mch, sta in pies.run( source=source, data=data ):
+                        pass
+                except cpppo.NonTerminal:
                     pass
-            except cpppo.NonTerminal:
-                pass
-
-            log.info( "%s ends: %s: %r", pies.name_centered(),
-                      "string accepted" if pies.terminal else "string rejected", data )
-            
-            assert pies.terminal == ( 'π' in text )
-            assert data.pies.input.tobytes().decode('utf-8') == text
+                accepted		= pies.terminal and data.pies.input.tobytes() == original
+                log.detail( "%s ends w/ re: %s: %s: %r", pies.name_centered(), re,
+                          "string accepted" if accepted else "string rejected", data )
+                expected		= tr == ('π' in text )
+                assert accepted == expected
+                assert original.startswith( data.pies.input.tobytes() )
 
 
 def test_decide():
@@ -537,7 +626,7 @@ def test_decide():
         for i,(m,s) in enumerate( comparo.run( source=source, data=data )):
             log.info( "%s #%3d -> %10.10s; next byte %3d: %-10.10r: %r", m.name_centered(),
                       i, s, source.sent, source.peek(), data )
-        assert i == 11
+        assert i == 12
         assert s is less
             
     source			= cpppo.peekable( str('a 33 33') )
@@ -546,7 +635,7 @@ def test_decide():
         for i,(m,s) in enumerate( comparo.run( source=source, data=data )):
             log.info( "%s #%3d -> %10.10s; next byte %3d: %-10.10r: %r", m.name_centered(),
                       i, s, source.sent, source.peek(), data )
-        assert i == 13
+        assert i == 14
         assert s is equal
 
 def test_limit():
@@ -586,7 +675,7 @@ def test_limit():
                      else data.odd_b.input.tounicode() ) == str( 'a'+'b'*9 )
         
 def test_decode():
-    # Test decode of regexes over bytes data.  Operates in raw bytes symbols.
+    # Test decode of regexes over bytes data.  Operates in raw bytes symbols., works in Python 2/3.
     source			= cpppo.peekable( 'π'.encode( 'utf-8' ))
     data			= cpppo.dotdict()
     with cpppo.string_bytes( 'pi', initial='.*', greedy=True, context='pi', decode='utf-8' ) as machine:
