@@ -199,12 +199,12 @@ logging.Logger.detail		= __detail
 
 def __normal_root( msg, *args, **kwargs ):
     if len( logging.root.handlers ) == 0:
-        basicConfig()
+        logging.basicConfig()
     logging.root.normal( msg, *args, **kwargs )
 
 def __detail_root( msg, *args, **kwargs ):
     if len( logging.root.handlers ) == 0:
-        basicConfig()
+        logging.basicConfig()
     logging.root.detail( msg, *args, **kwargs )
 
 change_function( __normal_root, co_filename=logging._srcfile )
@@ -254,14 +254,39 @@ def clamp( val, lim ):
 # decreasing ordering of domains or ranges.  If clamped, we will ensure that the rng is (re)ordered
 # appropriately.
 # 
-def scale( val, dom, rng, clamped=False ):
-    """Map 'val' from domain 'dom', to new range 'rng'"""
+#     If non-unity exponent is provided, then the input domain is raised to the appropriate power
+# during the mapping.  This allows us to map something like (25,40)->(0,1) with a curve such as:
+# 
+#   1 |              .
+#     |             .
+#     |           ..
+#     |        ...
+#     |   .....
+#   0 +---------------
+#     2              4
+#     5              0
+# 
+def scale( val, dom, rng, clamped=False, exponent=1 ):
+    """Map 'val' from domain 'dom', to new range 'rng', optionally with an exponential scaling.  If a
+    non-unity exponent is provided, then the input value is also clamped to the input domain (and
+    its order is asserted) since raising -'ve values to arbitrary exponents will usually have very
+    unexpected results.  Otherwise, at unity exponent, allow -'ve values and out-of-order ranges.
+
+    """
+    if exponent != 1:
+        assert dom[1] > dom[0], "Scaling %s non-linearly requires an ordered domain: %s" % ( val, dom )
+        if clamped:
+            val			= clamp( val, (min(dom),max(dom)) )
+        else:
+            assert dom[0] <= val <= dom[1], "Scaling %s non-linearly requires value in domain: %s" % ( val, dom )
+    else:
+        assert dom[1] != dom[0], "Scaling %s requires a non-zero domain: %s" % ( val, dom )
     result                      = ( rng[0]
-                                    + ( val    - dom[0] )
+                                    + ( val    - dom[0] ) ** exponent
                                     * ( rng[1] - rng[0] )
-                                    / ( dom[1] - dom[0] ))
+                                    / ( dom[1] - dom[0] ) ** exponent )
     if clamped:
-        result                  = clamp( result, (min(rng), max(rng)))
+        result                  = clamp( result, (min(rng),max(rng)))
     return result
 
 # 
