@@ -103,11 +103,13 @@ def octets_encode( value ):
 
 class octets_struct( octets_base, cpppo.state_struct ):
     """Scans octets sufficient to satisfy the specified struct 'format', and then parses it according
-    to the supplied struct 'format'."""
+    to the supplied struct 'format' (default is class-level struct_format attribute)."""
     def __init__( self, name=None, format=None, **kwds ):
-        assert isinstance( format, str ), "Expected a struct 'format', found: %r" % format
-        super( octets_struct, self ).__init__( name=name, repeat=struct.calcsize( format ),
-                                               format=format, **kwds )
+        if format is not None:
+            assert isinstance( format, cpppo.type_str_base ), "Expected a struct 'format', found: %r" % format
+        super( octets_struct, self ).__init__( name=name, format=format, 
+            repeat=struct.calcsize( self.struct_format if format is None else format ),
+                                               **kwds )
 
 
 class octets_noop( octets_base, cpppo.state ):
@@ -168,13 +170,15 @@ class words( words_base, cpppo.state ):
 #     More complex data types are derived from STRUCT, are derived from cpppo.dfa, and require a
 # state machine to be constructed to parse the data.
 # 
-#     Any EtherNet/IP type based on TYPE has a .calcsize property; its size in bytes.
+#     Any EtherNet/IP type based on TYPE has class-level .struct_format and a
+# .struct_calcsize attribute; its size in bytes (we do not employ the capability
+# of octets_struct and state_struct to have a custom instance-level format).
 # 
 class TYPE( octets_struct ):
     """An EtherNet/IP data type"""
     def __init__( self, name=None, **kwds ):
         name			= name or kwds.setdefault( 'context', self.__class__.__name__ )
-        super( TYPE, self ).__init__( name=name, format=self.struct_format, **kwds )
+        super( TYPE, self ).__init__( name=name, **kwds )
 
     @classmethod
     def produce( cls, value ):
@@ -182,38 +186,45 @@ class TYPE( octets_struct ):
 
 class USINT( TYPE ):
     """An EtherNet/IP USINT; 8-bit unsigned integer"""
-    tag_type			= None
+    tag_type			= 0x00c6
     struct_format		= 'B'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 class SINT( TYPE ):
     """An EtherNet/IP SINT; 8-bit signed integer"""
     tag_type			= 0x00c2
     struct_format		= 'b'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 class UINT( TYPE ):
     """An EtherNet/IP UINT; 16-bit unsigned integer"""
-    tag_type			= None
+    tag_type			= 0x00c7
     struct_format		= '<H'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 class INT( TYPE ):
     """An EtherNet/IP INT; 16-bit signed integer"""
     tag_type			= 0x00c3
     struct_format		= '<h'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 class UDINT( TYPE ):
     """An EtherNet/IP UDINT; 32-bit unsigned integer"""
-    tag_type			= None
+    tag_type			= 0x00c8
     struct_format		= '<I'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 class DINT( TYPE ):
     """An EtherNet/IP DINT; 32-bit signed integer"""
     tag_type			= 0x00c4
     struct_format		= '<i'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 class REAL( TYPE ):
     """An EtherNet/IP INT; 32-bit float"""
     tag_type			= 0x00ca
     struct_format		= '<f'
+    struct_calcsize		= struct.calcsize( struct_format )
 
 
 class STRUCT( cpppo.dfa, cpppo.state ):
@@ -352,8 +363,8 @@ def enip_format( data ):
 # EtherNet/IP CIP Parsing
 # 
 # See Vol2_1.14.pdf, Chapter 3-2.1 Unconnected Messages, for a simplified overview of parsing.  We
-# parse the SendRRData, the CPF encapsulation, and the CFP Null Address and [Un]connected Data
-# items, and finally the CIP Message Router Request from the second CFP item.
+# parse the SendRRData, the CPF encapsulation, and the CPF Null Address and [Un]connected Data
+# items, and finally the CIP Message Router Request from the second CPF item.
 # 
 
 class move_if( cpppo.decide ):
@@ -377,20 +388,20 @@ class move_if( cpppo.decide ):
                                     if not hasattr( self.ini, '__call__' )
                                     else self.ini(
                                             machine=machine, source=source, path=path, data=data ))
-                log.debug( "%s -- init. data[%r] to %r in data: %s", self, pathdst, ini, data )
-                data[pathdst]	= ini
+                try:
+                    data[pathdst]= ini
+                finally:
+                    log.debug( "%s -- init. data[%r] to %r in data: %s", self, pathdst, ini, data )
             if self.src is not None:
                 pathsrc		= path + self.src
                 assert pathsrc in data, \
                     "Could not find %r to move to %r in %r" % ( pathsrc, pathdst, data )
                 if hasattr( data[pathdst], 'append' ):
-                    log.debug( "%s -- append data[%r] == %r to data[%r]", self,
-                               pathsrc, data[pathsrc], pathdst )
+                    log.debug( "%s -- append data[%r] == %r to data[%r]", self, pathsrc, data[pathsrc], pathdst )
                     data[pathdst].append( data.pop( pathsrc ))
                 else:
-                    log.debug( "%s -- assign data[%r] == %r to data[%r]", self,
-                               pathsrc, data[pathsrc], pathdst )
-                    data[pathdst]	= data.pop( pathsrc )
+                    log.debug( "%s -- assign data[%r] == %r to data[%r]", self, pathsrc, data[pathsrc], pathdst )
+                    data[pathdst] = data.pop( pathsrc )
 
         return target
 
@@ -410,16 +421,79 @@ class EPATH( cpppo.dfa ):
          .EPATH.segment__... temp 
 
     Also works as a Route Path (which has a pad after size), by setting padsize=True.
-    """
-    padsize			= False
 
+    The path logical segments are encoded as follows (from Volume 1: CIP Common Specification,
+    C=1.4.2 Logical Segment):
+
+
+    Segment Type    Logical Type    Logical Format
+    +---+---+---+   +---+---+---+   +---+---+
+    | 0 | 0 | 1 |   |   |   |   |   |   |   |
+    +---+---+---+   +---+---+---+   +---+---+
+
+    Class ID          0   0   0       0   0   8-bit logical address
+    Instance ID       0   0   1       0   1  16-bit logical address
+    Element/Member ID 0   1   0       1   0  32-bit logical address (element Id's only?)
+    Connection Point  0   1   1       1   1  (reserved for future use)
+    Attribute ID      1   0   0
+    Special*          1   0   1
+    Service ID*       1   1   0
+    Reserved          1   1   1
+
+    *The Special and Service ID Logical Types do not use the logical addressing definition for the Logical Format
+
+    The 8-bit logical address format is allowed for use with all Logical Types.
+
+    The 16-bit logical address format is only allowed for use with Logical Types Class ID,
+    Instance ID, Member ID, and Connection Point.
+
+    The 32-bit logical address format is not allowed (reserved for future use).
+
+    The Connection Point Logical Type provides additional addressing capabilities beyond the
+    standard Class ID/Instance ID/Attribute ID/Member ID Object Address. Object Classes shall
+    define when and how this addressing component is utilized.
+
+    The Service ID Logical Type has the following definition for the Logical Format:
+        0 0 8-Bit Service ID Segment (0x38)
+        0 1 Reserved for future use (0x39)
+        1 0 Reserved for future use (0x3A)
+        1 1 Reserved for future use (0x3B)
+
+    The Special Logical Type has the following definition for the Logical Format:
+        0 0 Electronic Key Segment (0x34)
+        0 1 Reserved for future use (0x35)
+        1 0 Reserved for future use (0x36)
+        1 1 Reserved for future use (0x37)
+
+    The Electronic Key segment shall be used to verify/identify a device. Possible uses include
+    verification during connection establishment and identification within an EDS file. This segment
+    has the format as shown in the table below.
+
+
+    From Volume 1, Common Industrial Protocol Specification, 5-5.5.3:
+
+    Connection Points within the Assembly Object are identical to Instances. For example, Connection
+    Point 4 of the Assembly Object is the same as Instance 4. Specifying a path of "20 04 24 VV 30
+    03" is the same as "20 04 2C VV 30 03".
+
+    """
+    PADSIZE			= False
+    SEGMENTS			= {
+        'symbolic':	0x91,
+        'class':	0x20,
+        'instance':	0x24,
+        'connection':	0x2c, # In Assembly Class 0x0004, is like instance
+        'attribute':	0x30,
+        'element':	0x28,
+        'port':		0x00,
+    }
     def __init__( self, name=None, **kwds ):
         name 			= name or kwds.setdefault( 'context', self.__class__.__name__ )
 
         # Get the size, and chain remaining machine onto rest.  When used as a Route Path, the size
         # is padded, so insert a state to drop the pad, and chain rest to that instead.
         size		= rest	= USINT(			context='size' )
-        if self.padsize:
+        if self.PADSIZE:
             size[True]	= rest	= octets_drop( 	'pad', 		repeat=1 )
 
         # After capturing each segment__ (pseg), move it onto the path segment list, and loop
@@ -430,6 +504,10 @@ class EPATH( cpppo.dfa ):
                                                 state=pseg )
 
         # Wire each different segment type parser between pseg and pmov
+        
+        # 0x28 == 001 010 00 Class ID,  8-bit
+        # 0x29 == 001 010 01 Class ID, 16-bit
+        # 0x2A == 001 010 11 Class ID, 32-bit
         pseg[b'\x28'[0]]= e_8t	= octets_drop(	'type',		repeat=1 )
         e_8t[True]	= e_8v	= USINT( 	'elem_8bit',	context='element')
         e_8v[None]		= pmov
@@ -442,7 +520,8 @@ class EPATH( cpppo.dfa ):
         e32t[True]	= e32v	= UDINT(	'elem32bit',	context='element')
         e32v[None]		= pmov
 
-
+        # 0x20 == 001 000 00 Class ID,  8-bit
+        # 0x21 == 001 000 01 Class ID, 16-bit
         pseg[b'\x20'[0]]= c_8t	= octets_drop(	'type',		repeat=1 )
         c_8t[True]	= c_8v	= USINT(	'clas_8bit',	context='class')
         c_8v[None]		= pmov
@@ -451,7 +530,8 @@ class EPATH( cpppo.dfa ):
         c16t[True]	= c16v	= UINT(		'clas16bit',	context='class')
         c16v[None]		= pmov
 
-
+        # 0x24 == 001 001 00 Instance ID,  8-bit
+        # 0x25 == 001 001 01 Instance ID, 16-bit
         pseg[b'\x24'[0]]= i_8t	= octets_drop(	'type',		repeat=1 )
         i_8t[True]	= i_8v	= USINT(	'inst_8bit',	context='instance')
         i_8v[None]		= pmov
@@ -460,7 +540,18 @@ class EPATH( cpppo.dfa ):
         i16t[True]	= i16v	= UINT(		'inst16bit',	context='instance')
         i16v[None]		= pmov
 
+        # 0x2C == 001 011 00 Connection Point,  8-bit
+        # 0x2D == 001 011 01 Connection Point, 16-bit
+        pseg[b'\x2c'[0]]= p_8t	= octets_drop(	'type',		repeat=1 )
+        p_8t[True]	= p_8v	= USINT(	'cnpt_8bit',	context='connection')
+        p_8v[None]		= pmov
 
+        pseg[b'\x2d'[0]]= p16t	= octets_drop(	'type',		repeat=2 )
+        p16t[True]	= p16v	= UINT(		'cnpt16bit',	context='connection')
+        p16v[None]		= pmov
+
+        # 0x30 == 001 100 00 Attribute ID,  8-bit
+        # 0x31 == 001 100 01 Attribute ID, 16-bit
         pseg[b'\x30'[0]]= a_8t	= octets_drop(	'type',		repeat=1 )
         a_8t[True]	= a_8v	= USINT(	'attr_8bit',	context='attribute')
         a_8v[None]		= pmov
@@ -469,7 +560,7 @@ class EPATH( cpppo.dfa ):
         a16t[True]	= a16v	= UINT(		'attr16bit',	context='attribute')
         a16v[None]		= pmov
 
-
+        # 0x90 == 100 100 01 Symbolic
         pseg[b'\x91'[0]]= symt	= octets_drop(	'type',		repeat=1 )
         symt[True]	= syml	= USINT(	'sym_len',	context='symbolic.length' )
         syml[None]	= symv	= cpppo.string_bytes(
@@ -600,13 +691,7 @@ class EPATH( cpppo.dfa ):
         result			= b''
         for seg in data.segment:
             found			= False
-            for segnam, segtyp in {
-                    'symbolic':		0x91,
-                    'class':		0x20,
-                    'instance':		0x24,
-                    'attribute':	0x30,
-                    'element':		0x28,
-                    'port':		0x00, }.items():
+            for segnam, segtyp in cls.SEGMENTS.items():
                 if segnam not in seg:
                     continue
                 found		= True
@@ -647,7 +732,7 @@ class EPATH( cpppo.dfa ):
                             result += b'\00'
                     break
 
-                # A numeric path segment; class, instance, attribute, element:
+                # A numeric path segment; class, instance/connection, attribute, element:
                 if segval <= 0xff:
                     result     += USINT.produce( segtyp )
                     result     += USINT.produce( segval )
@@ -668,7 +753,7 @@ class EPATH( cpppo.dfa ):
             assert len( result ) % 2 == 0, \
                 "Failed to retain even EPATH word length after %r in %r" % ( segnam, data )
     
-        return USINT.produce( len( result ) // 2 ) + ( b'\x00' if cls.padsize else b'' ) + result
+        return USINT.produce( len( result ) // 2 ) + ( b'\x00' if cls.PADSIZE else b'' ) + result
 
 
 class route_path( EPATH ):
@@ -679,7 +764,7 @@ class route_path( EPATH ):
         .route_path.segment 		...
 
     """
-    padsize			= True
+    PADSIZE			= True
 
 
 class unconnected_send( cpppo.dfa ):
@@ -760,7 +845,60 @@ class unconnected_send( cpppo.dfa ):
         return result
 
 
+class communications_service( cpppo.dfa ):
+    """The ListServices response contains a CPF item list containing one item: a "Communications"
+    type_id 0x0100, indicating that the device supports encapsulation of CIP packets.  These CPF
+    items contain the standard type_id and length, followed by:
+
+       .CPF.item[0].version		UINT		2	Version of protocol (shall be 1)
+       .CPF.item[0].capability		UINT		2	Capability flags
+       .CPF.item[0].service_name	USINT[*]    .length-8	Name of service + NUL (eg. "Communications\0")
+
+    +-------------+---------------------------------------------------------------+
+    | Flag        | Description                                                   |
+    +-------------+---------------------------------------------------------------+
+    | Bits 0 - 4  | Reserved for legacy usage 1                                   |
+    | Bit 5       | If the device supports EtherNet/IP encapsulation of CIP       |
+    |             | this bit shall be set (=1); otherwise, it shall be clear (=0) |
+    | Bits 6 - 7  | Reserved for legacy usage 1                                   |
+    | Bit 8       | Supports CIP transport class 0 or 1 UDP-based connections     |
+    | Bits 9 - 15 | Reserved for future expansion                                 |
+    +-------------+---------------------------------------------------------------+
+
+
+    """
+    def __init__( self, name=None, **kwds ):
+        name 			= name or kwds.setdefault( 'context', self.__class__.__name__ )
+        
+        vers			= UINT(	context='version' )
+        vers[True]	= capa	= UINT(	context='capability' )
+
+        capa[True]	= svnm	= cpppo.string_bytes( 'service_name',
+                                        context='service_name', greedy=True,
+                                        initial='[^\x00]*', decode='iso-8859-1' )
+        svnm[b'\0'[0]]		= octets_drop( 'NUL', repeat=1, terminal=True )
+
+        '''
+        capa[b'\0'[0]]	= done	= octets_drop( 'NUL', repeat=1, terminal=True )
+        capa[True]	= svnm	= octets(	context='service_name' )
+        svnm[b'\0'[0]]	= done
+        svnm[True]	= svnm
+        '''
+
+        super( communications_service, self ).__init__( name=name, initial=vers, **kwds )
+
+    @classmethod
+    def produce( cls, data ):
+        result			= b''
+        result	       	       += UINT.produce( data.version )
+        result	               += UINT.produce( data.capability )
+        result		       += data.service_name.encode( 'iso-8859-1' )
+        result		       += b'\0'
+        return result
+
+
 class CPF( cpppo.dfa ):
+
     """A SendRRData Common Packet Format specifies the number and type of the encapsulated CIP
     address items or data items that follow:
 
@@ -785,17 +923,23 @@ class CPF( cpppo.dfa ):
         0x0100:		ListServices response
 
     
-    Presently we only handle NULL Address and Unconnected Messages.
+    Presently we only handle NULL Address and Unconnected Messages, and ListServices.
 
     """
-    item_parsers		= {
-            0x00b2: 	unconnected_send,
-    } 
+    ITEM_PARSERS		= {
+            0x00b2:	unconnected_send,	# used in SendRRData request/response
+            0x0100:	communications_service, # used in ListServices response
+    }
 
     def __init__( self, name=None, **kwds ):
         """Parse CPF list items 'til .count reached, which should be simultaneous with symbol exhaustion, if
         caller specified a symbol limit.
 
+        A CPF list may be completely empty (ie. not even a 0 count), for certain use-cases.  For
+        example, a EtherNet/IP CIP ListServices request consists solely of a CIP frame consisting of
+        a CPF list containing a Communications Service item.  However, the request is simply missing
+        the CPF list -- completely.  So, make the initial state produce an empty CPF dotdict.
+
         """
         name 			= name or kwds.setdefault( 'context', self.__class__.__name__ )
 
@@ -808,10 +952,11 @@ class CPF( cpppo.dfa ):
 
         # Prepare a parser for each recognized CPF item type.  It must establish one level of
         # context, because we need to pass it a limit='..length' denoting the length we just parsed.
-
-        for typ,cls in ( self.item_parsers or {} ).items():
+        # Note that we must capture the value of 'typ' in the lambda definition as a keyword
+        # parameter (which is evaluated at once), or it will take the final value of outer 'typ'
+        for typ,cls in self.ITEM_PARSERS.items():
             ilen[None]		= cpppo.decide( cls.__name__, state=cls( terminal=True, limit='..length' ),
-                        predicate=lambda path=None, data=None, **kwds: data[path].type_id == typ )
+                        predicate=lambda path=None, data=None, typ=typ, **kwds: data[path].type_id == typ )
 
         # If we don't recognize the CPF item type, just parse remainder into .input (so we could re-generate)
         ilen[None]	= urec	= octets( 	'unrecognized',	context=None,
@@ -826,23 +971,27 @@ class CPF( cpppo.dfa ):
                                            destination='.item', initializer=lambda **kwds: [] )
         item[None]		= cpppo.state( 	'done', terminal=True )
 
-        # Parse count, and then exactly .count CPF items.
-        loop			= UINT( 			context='count' )
+        # Parse count, and then exactly .count CPF items (or just an empty dict, if nothing)
+        emty			= octets_noop(	'empty',	terminal=True )
+        emty.initial[None]	= move_if( 	'mark',		initializer={} )
+        emty[True]	= loop	= UINT( 			context='count' )
         loop[None]		= cpppo.dfa( 	'all', 	
                                                 initial=item,	repeat='.count',
                                                 terminal=True )
 
-        super( CPF, self ).__init__( name=name, initial=loop, **kwds )
+        super( CPF, self ).__init__( name=name, initial=emty, **kwds )
 
     @classmethod
     def produce( cls, data ):
         """Regenerate a CPF message structure.   """
         result			= b''
+        if not data:
+            return result # An empty CPF -- indicates no CPF segment present at all
         result		       += UINT.produce( len( data.item ))
         for item in data.item:
             result	       += UINT.produce( item.type_id )
-            if item.type_id in cls.item_parsers:
-                itmprs		= cls.item_parsers[item.type_id] # eg 'unconnected_send'
+            if item.type_id in cls.ITEM_PARSERS:
+                itmprs		= cls.ITEM_PARSERS[item.type_id] # eg 'unconnected_send', 'communications_service'
                 item.input	= bytearray( itmprs.produce( item[itmprs.__name__] ))
             if 'input' in item:
                 result	       += UINT.produce( len( item.input ))
@@ -851,6 +1000,7 @@ class CPF( cpppo.dfa ):
                 result	       += UINT.produce( 0 )
         return result
 
+
 class send_data( cpppo.dfa ):
     """Handle Connected (SendUnitData) or Unconnected (SendRRData) Send Data request/reply."""
     def __init__( self, name=None, **kwds ):
@@ -895,7 +1045,7 @@ class unregister( octets_noop ):
     def terminate( self, exception, machine=None, path=None, data=None ):
         """Just create an empty value to indicate the command was received."""
         # Only operate if we have completed without exception.
-        super( state_struct, self ).terminate(
+        super( unregister, self ).terminate(
             exception=exception, machine=machine, path=path, data=data )
         if exception is not None:
             return
@@ -904,7 +1054,12 @@ class unregister( octets_noop ):
 
 
 class list_services( cpppo.dfa ):
-    """Handle ListServices request.  Services are encoded as a CPF list."""
+    """Handle ListServices request/reply.  Services are encoded as a CPF list.  We must deduce whether
+    we are parsing a request or a reply.  The request will have a 0 length; the reply (which must
+    contain a CPF with at least an item count) will have a non-zero length.
+
+    Even if the request is empty, we want to produce 'CIP.list_services.CPF'.
+    """
     def __init__( self, name=None, **kwds ):
         name 			= name or kwds.setdefault( 'context', self.__class__.__name__ )
 
@@ -912,6 +1067,13 @@ class list_services( cpppo.dfa ):
 
         super( list_services, self ).__init__( name=name, initial=svcs, **kwds )
 
+    @staticmethod
+    def produce( data ):
+        result			= b''
+        if 'CPF' in data:
+            result	       += CPF.produce( data.CPF )
+        return result
+
 
 class CIP( cpppo.dfa ):
     """The EtherNet/IP CIP Encapsulation transports various commands back and forth between a
@@ -949,7 +1111,7 @@ class CIP( cpppo.dfa ):
         .CIP.unregister
 
     ListServices		0x0004
-        .CIP.listservices.CPF...	...
+        .CIP.list_services.CPF...	...
 
     SendRRData			0x006f
     SendUnitData		0x0070
@@ -958,27 +1120,24 @@ class CIP( cpppo.dfa ):
         .CIP.send_data.CPF...
 
     """
+    COMMAND_PARSERS		= {
+        (0x006f,0x0070):	send_data,	# 0x006f (SendRRData) is default if CIP.send_data seen
+        (0x0065,):		register,
+        (0x0066,):		unregister,
+        (0x0004,):		list_services,
+    }
     def __init__( self, name=None, **kwds ):
         name 			= name or kwds.setdefault( 'context', self.__class__.__name__ )
 
         slct			= octets_noop(	'select' )
-        slct[None]		= cpppo.decide( 'SendData',
-            predicate=lambda path=None, data=None, **kwds: data[path+'..command'] in (0x006f,0x0070),
-                                            state=send_data(	limit='...length', terminal=True ))
-        slct[None]		= cpppo.decide( 'Register', 
-            predicate=lambda path=None, data=None, **kwds: data[path+'..command'] == 0x0065,
-                                            state=register(	limit='...length', terminal=True ))
-        slct[None]		= cpppo.decide( 'Unregister',
-            predicate=lambda path=None, data=None, **kwds: data[path+'..command'] == 0x0066,
-                                            state=unregister(	limit='...length', terminal=True ))
-        slct[None]		= cpppo.decide( 'ListServices',
-            predicate=lambda path=None, data=None, **kwds: data[path+'..command'] == 0x0004,
-                                            state=list_services(limit='...length', terminal=True ))
-
+        for cmd,cls in self.COMMAND_PARSERS.items():
+            slct[None]		= cpppo.decide( cls.__name__,
+                    state=cls( limit='...length', terminal=True ),
+                    predicate=lambda path=None, data=None, cmd=cmd, **kwds: data[path+'..command'] in cmd )
         super( CIP, self ).__init__( name=name, initial=slct, **kwds )
 
-    @staticmethod
-    def produce( data ):
+    @classmethod
+    def produce( cls, data ):
         """Expects to find a recognized .command value and/or and parsed .CIP.register,
         .CIP.unregister, .etc. in the provided data artifact as produced by our parser.  Produces
         the bytes string encoding the command.  There is little difference between a request and a
@@ -986,23 +1145,16 @@ class CIP( cpppo.dfa ):
         response it may indicate an error.
 
         """
-        result			= b''
-        
-        # TODO: handle remaining requests; unregister, ...
-        if ( data.get( 'command' ) == 0x0065
-             or 'CIP.register' in data and data.setdefault( 'command', 0x0065 ) == 0x0065 ):
-            result	       += register.produce( data.CIP.register )
-        elif ( data.get( 'command' ) == 0x006f
-               or 'CIP.send_data' in data and  data.setdefault( 'command', 0x006f ) == 0x006f ):
-            result	       += send_data.produce( data.CIP.send_data )
-        else:
-            assert False, "Invalid CIP request/reply format: %r" % data
-
-        return result
+        for cmd,cmdcls in cls.COMMAND_PARSERS.items():
+            if ( data.get( 'command' ) in cmd
+                 or ( 'CIP.' + cmdcls.__name__  in data
+                      and data.setdefault( 'command', cmd[0] ) in cmd )):
+                return cmdcls.produce( data['CIP.' + cmdcls.__name__] )
+        raise Exception( "Invalid CIP request/reply format: %r" % data )
 
 class typed_data( cpppo.dfa ):
     """Parses CIP typed data, of the form specified by the datatype (must be a relative path within
-    the data artifact).  Data elements are parsed 'til exhaustion of input, so the caller should
+    the data artifact, or an integer data type).  Data elements are parsed 'til exhaustion of input, so the caller should
     use limit= to define the limits of the data in the source symbol input stream; only complete
     data items must be parsed, so this must be exact, and match the specified data type.
 
@@ -1015,13 +1167,26 @@ class typed_data( cpppo.dfa ):
     INT		yes		= 0x00c3	# 2 bytes
     DINT	yes		= 0x00c4	# 4 bytes
     REAL	yes		= 0x00ca	# 4 bytes
+    USINT	yes		= 0x00c6	# 1 byte
+    UINT	yes		= 0x00c7	# 2 bytes
+    UDINT	yes		= 0x00c8	# 4 bytes
     DWORD			= 0x00d3	# 4 byte (32-bit boolean array)
     LINT			= 0x00c5	# 8 byte
 
     """
-    def __init__( self, name=None, datatype=None, **kwds ):
+    TYPES_SUPPORTED		= {
+        SINT.tag_type:	SINT,
+        USINT.tag_type:	USINT,
+        INT.tag_type:	INT,
+        UINT.tag_type:	UINT,
+        DINT.tag_type:	DINT,
+        UDINT.tag_type:	UDINT,
+        REAL.tag_type:	REAL,
+    }
+
+    def __init__( self, name=None, tag_type=None, **kwds ):
         name 			= name or kwds.setdefault( 'context', self.__class__.__name__ )
-        assert datatype, "Must specify a relative path to the CIP data type; found: %r" % datatype
+        assert tag_type, "Must specify a numeric (or relative path to) the CIP data type; found: %r" % tag_type
 
         slct			= octets_noop(	'select' )
         
@@ -1032,6 +1197,13 @@ class typed_data( cpppo.dfa ):
                                            destination='.data',	initializer=lambda **kwds: [],
                                                 state=i_8d )
 
+        u_8d			= octets_noop(	'end_8bitu',
+                                                terminal=True )
+        u_8d[True]	= u_8p	= USINT()
+        u_8p[None]		= move_if( 	'mov_8bitu',	source='.USINT', 
+                                           destination='.data',	initializer=lambda **kwds: [],
+                                                state=u_8d )
+
         i16d			= octets_noop(	'end16bit',
                                                 terminal=True )
         i16d[True]	= i16p	= INT()
@@ -1039,6 +1211,13 @@ class typed_data( cpppo.dfa ):
                                            destination='.data',	initializer=lambda **kwds: [],
                                                 state=i16d )
 
+        u16d			= octets_noop(	'end16bitu',
+                                                terminal=True )
+        u16d[True]	= u16p	= UINT()
+        u16p[None]		= move_if( 	'mov16bitu',	source='.UINT', 
+                                           destination='.data',	initializer=lambda **kwds: [],
+                                                state=u16d )
+
         i32d			= octets_noop(	'end32bit',
                                                 terminal=True )
         i32d[True]	= i32p	= DINT()
@@ -1046,6 +1225,13 @@ class typed_data( cpppo.dfa ):
                                            destination='.data',	initializer=lambda **kwds: [],
                                                 state=i32d )
 
+        u32d			= octets_noop(	'end32bitu',
+                                                terminal=True )
+        u32d[True]	= u32p	= UDINT()
+        u32p[None]		= move_if( 	'mov32bitu',	source='.UDINT', 
+                                           destination='.data',	initializer=lambda **kwds: [],
+                                                state=u32d )
+
         fltd			= octets_noop(	'endfloat',
                                                 terminal=True )
         fltd[True]	= fltp	= REAL()
@@ -1054,42 +1240,46 @@ class typed_data( cpppo.dfa ):
                                                 state=fltd )
 
         slct[None]		= cpppo.decide(	'SINT',	state=i_8p,
-            predicate=lambda path=None, data=None, **kwds: data[path+datatype] == SINT.tag_type )
+            predicate=lambda path=None, data=None, **kwds: \
+                SINT.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
+        slct[None]		= cpppo.decide(	'USINT',state=u_8p,
+            predicate=lambda path=None, data=None, **kwds: \
+                USINT.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
         slct[None]		= cpppo.decide(	'INT',	state=i16p,
-            predicate=lambda path=None, data=None, **kwds: data[path+datatype] == INT.tag_type )
+            predicate=lambda path=None, data=None, **kwds: \
+                INT.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
+        slct[None]		= cpppo.decide(	'UINT',	state=u16p,
+            predicate=lambda path=None, data=None, **kwds: \
+                UINT.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
         slct[None]		= cpppo.decide(	'DINT',	state=i32p,
-            predicate=lambda path=None, data=None, **kwds: data[path+datatype] == DINT.tag_type )
+            predicate=lambda path=None, data=None, **kwds: \
+                DINT.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
+        slct[None]		= cpppo.decide(	'UDINT',state=u32p,
+            predicate=lambda path=None, data=None, **kwds: \
+                UDINT.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
         slct[None]		= cpppo.decide(	'REAL',	state=fltp,
-            predicate=lambda path=None, data=None, **kwds: data[path+datatype] == REAL.tag_type )
+            predicate=lambda path=None, data=None, **kwds: \
+                REAL.tag_type == ( data[path+tag_type] if isinstance( tag_type, cpppo.type_str_base ) else tag_type ))
         
         super( typed_data, self ).__init__( name=name, initial=slct, **kwds )
 
-    @staticmethod
-    def produce( data ):
-        """Expects to find .type and .data.list, and produces the data encoded to bytes."""
-        assert 'type' in data and data.type in ( SINT.tag_type, INT.tag_type, DINT.tag_type, REAL.tag_type ), \
-            "Unknown (or no) .type found: %r" % data
-        if data.type ==   SINT.tag_type:
-            return b''.join(  SINT.produce( v ) for v in data.data )
-        elif data.type ==  INT.tag_type:
-            return b''.join(   INT.produce( v ) for v in data.data )
-        elif data.type == DINT.tag_type:
-            return b''.join(  DINT.produce( v ) for v in data.data )
-        elif data.type == REAL.tag_type:
-            return b''.join(  REAL.produce( v ) for v in data.data )
+    @classmethod
+    def produce( cls, data, tag_type=None ):
+        """Expects to find .type (if tag_type is None) and .data list, and produces the data encoded to bytes."""
+        if tag_type is None:
+            tag_type		= data.get( 'type' )
+        assert hasattr( data, '__iter__' ) and 'data' in data and tag_type in cls.TYPES_SUPPORTED, \
+            "Unknown (or no) typed data found for tag_type %r: %r" % ( tag_type, data )
+        produce			= cls.TYPES_SUPPORTED[tag_type].produce
+        return b''.join( produce( v ) for v in data.data )
+
+    @classmethod
+    def datasize( cls, tag_type, size=1 ):
+        """Compute the encoded data size for the specified tag_type and amount of data."""
+        assert tag_type in cls.TYPES_SUPPORTED, \
+            "Unknown tag_type %r" % ( tag_type )
+        return cls.TYPES_SUPPORTED[tag_type].struct_calcsize * size
 
-    @staticmethod
-    def estimate( tag_type, size ):
-        """Compute the encoded data size for the specified type and amount of data."""
-        if tag_type ==   SINT.tag_type:
-            return 1 * size
-        elif tag_type ==  INT.tag_type:
-            return 2 * size
-        elif tag_type == DINT.tag_type:
-            return 4 * size
-        elif tag_type == REAL.tag_type:
-            return 4 * size
-        
 
 class status( cpppo.dfa ):
     """Parses CIP status, and status_ext.size/.data:
